from functools import lru_cache
from typing import List
import typer
from typing_extensions import Annotated, Optional
from crap.crap_manager import CrapManager
from crap.package_management import PackageManagement

__all__: List[str] = ["get_app"]

app = typer.Typer(no_args_is_help=True)


@app.command()
def crap(
    path_: Annotated[str, typer.Argument(help="path to file/files")] = ".",
    important: Optional[str] = typer.Option(
        None,
        "--important",
        "-i",
        help="Add a package to the list of important packages",
    ),
    remove: Optional[str] = typer.Option(
        None,
        "--remove",
        "-r",
        help="Remove a package from the list of important packages",
    ),
    flush: bool = typer.Option(
        False,
        "--flush",
        "-f",
        help="Remove all packages from the list of important packages",
    ),
    show: bool = typer.Option(
        False,
        "--show",
        "-s",
        help="Show all important packages",
    ),
    factory_reset: bool = typer.Option(
        False,
        "--factory-reset",
        "-fr",
        help="Reset all settings to default",
    ),
):
    if (
        sum(
            [
                bool(opt)
                for opt in [path_ != ".", important, remove, flush, show, factory_reset]
            ]
        )
        > 1
    ):
        print("Error: Options cannot be used together.")
        raise typer.Exit(code=1)

    package_management = PackageManagement()
    if important:
        package_management.add_important_package(important)
    elif remove:
        package_management.remove_important_package(remove)
    elif flush:
        package_management.flush_important_packages()
    elif show:
        package_management.show_important_packages()
    elif factory_reset:
        package_management.factory_reset_important_packages()
    else:
        manager = CrapManager(path_=path_)
        manager.run()


@lru_cache
def get_app():
    """
    Get the main typer cli app
    """
    return app()



import subprocess
import json
from functools import lru_cache
from typing import List, Dict
import pathlib


def get_current_packages() -> set:
    """Get the current packages installed in the environment."""
    process = subprocess.run(["pip3", "freeze"], capture_output=True, text=True)
    output = process.stdout.strip()
    packages = set(line.split("==")[0] for line in output.split("\n"))
    return packages


def run_subprocess(command: list[str]) -> subprocess.CompletedProcess:
    """Run a given subprocess command."""
    try:
        return subprocess.run(command, capture_output=True, text=True)
    except Exception as ex:
        print(ex)
        return None


def execute_command_without_output(command: list[str]):
    """Execute a command without needing the output."""
    run_subprocess(command)


def get_package_counter_dict() -> Dict[str, int]:
    """Get a dictionary of package usage counters."""
    packages_ = get_current_packages()
    return {package: 0 for package in packages_}


def freeze_into_requirements():
    """Freeze current environment to requirements.txt."""
    req_path = pathlib.Path(__file__).parent / "data" / "req.txt"
    try:
        with open(req_path, "w") as file_:
            subprocess.run(["pip3", "freeze"], stdout=file_)
    except Exception as ex:
        print(ex)


def reinstall_from_requirements() -> None:
    """Reinstall packages from requirements.txt."""
    req_path = pathlib.Path(__file__).parent / "data" / "req.txt"
    execute_command_without_output(
        ["pip3", "install", "-r", req_path, "--no-cache-dir"]
    )


def uninstall_package(package_name: str):
    """Uninstall a given package."""
    execute_command_without_output(["pip3", "uninstall", "-y", package_name])


def pre_cleanup_with_ruff(path_):
    """
    Pre cleanup with ruff
    """
    execute_command_without_output(["ruff", "check", path_, "--fix"])



import os
from typing import Set
import pathlib
from crap.file_analyzer import PythonFileAnalyzer
from crap.virtual_env_checker import VirtualEnvChecker
from crap.package_usage_counter import PackageUsageCounter
from crap.subprocesses import (
    uninstall_package,
    pre_cleanup_with_ruff,
    reinstall_from_requirements,
    freeze_into_requirements,
    get_current_packages
)


class CrapManager:
    __slots__ = ("path_", "venv_checker", "package_usage_counter", "deleted_packages")

    def __init__(self, path_: str):
        self.path_ = pathlib.Path(path_).absolute()
        self.venv_checker = VirtualEnvChecker()
        self.package_usage_counter = PackageUsageCounter()
        self.deleted_packages = set()
    
    def run(self):
        if not self.path_.exists():
            raise FileNotFoundError("File/Dir not found")

        self._process_path()
        initial_packages = get_current_packages()
        self._cleanup_packages()
        # After we cleanup remove the unused packages, we need to freeze the current environment to requirements.txt.
        # The left packages in requirements.txt are the ones that are actually used by the project. These packages might have dependecies
        # that we have deleted since they haven't directly been used by the project. So we need to reinstall the packages from requirements.txt.
        # When using pip to install packages, any dependecy gets installed automatically. So we don't need to worry about that.
        freeze_into_requirements()
        reinstall_from_requirements()

        final_packages = get_current_packages()
        self.deleted_packages = initial_packages - final_packages

        self.print_deleted_packages()
        

    def _process_path(self):
        if self.path_.is_file():
            self._analyze_file(self.path_)
        elif self.path_.is_dir():
            self._analyze_directory()

    def _analyze_file(self, file_path):
        pre_cleanup_with_ruff(file_path)
        analyzer = PythonFileAnalyzer(file_path)
        analyzer.analyze()
        for package in get_current_packages():
            if package in analyzer.imported_modules:
                self.package_usage_counter.increment_package_count(package)

    def _analyze_directory(self):
        """
        Analyzes the directory and its subdirectories, excluding certain directories,
        and analyzes each Python file found.

        This method walks through the directory tree using os.walk() function,
        excluding directories that are likely to be virtual environments or specified
        in the excluded_dirs list. For each file with a ".py" extension, it calls
        the _analyze_file() method to perform analysis.

        Parameters:
            None

        Returns:
            None
        """
        excluded_dirs = self._get_excluded_dirs()

        for dirpath, dirnames, filenames in os.walk(self.path_):
            dirnames[:] = [
                d
                for d in dirnames
                if not self.venv_checker.is_likely_venv(os.path.join(dirpath, d))
                and d not in excluded_dirs
            ]

            for filename in filenames:
                if filename.endswith(".py"):
                    file_path = pathlib.Path(dirpath) / filename
                    self._analyze_file(file_path)

    def _cleanup_packages(self):
        important_packages = self._get_important_packages()
        unused_packages = self.package_usage_counter.get_unused_packages(
            important_packages
        )

        for package in unused_packages:
            uninstall_package(package)
            self.deleted_packages.add(package)

    @staticmethod
    def _get_important_packages() -> Set[str]:
        """
        Retrieves the important packages from the important_packages.txt file.
        These packages cannot be removed, because they are most likely dev tools that do not
        get imported in the code.

        Returns:
            A set of strings representing the important packages.
        """
        with open(
            f"{pathlib.Path(__file__).parent}/data/important_packages.txt",
            "r",
            encoding="utf-8",
        ) as file:
            return {line.strip() for line in file}

    def print_deleted_packages(self):
        print("🗑️📦Deleted packages:")
        for package in self.deleted_packages:
            print(f" - {package}")

    @staticmethod
    def _get_excluded_dirs() -> Set[str]:
        """
        These directories are excluded from the analysis.
        """
        return {"__pycache__", ".git"}



from typing import List
from crap.subprocesses import get_package_counter_dict


class PackageUsageCounter:
    def __init__(self):
        self.pack_counter = get_package_counter_dict()

    def increment_package_count(self, package):
        if package in self.pack_counter:
            self.pack_counter[package] += 1

    def get_unused_packages(self, important_packages) -> List[str]:
        """
        Returns a list of unused packages.

        A package is considered unused if its count is 0 and it is not in the list of important packages.
        
        Args:
            important_packages (List[str]): A list of important packages.

        Returns:
            List[str]: A list of unused packages.
        """
        return [pkg for pkg, count in self.pack_counter.items() if count == 0 and pkg not in important_packages]
